/*
 * WINplot.c
 *      written by Masashi Katsurada */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#define G_DOUBLE
#define DOUBLE
#include <glsc.h>

#include "MyBuffer.h"
#include "fplot.h"

static void current_window_size(unsigned int *, unsigned int *);
static void writeshort(int, FILE *);
static int winx(int);
static int winy(int);
static int plx(int);
static int ply(int);
static int toasciz(char *);
static void deltailnewline(char *);
static void deltailsp(char *);

#ifndef M_PI
#define PI      3.14159265358979323846
#else
#define PI      M_PI
#endif

/* 饸٤Ѵ (ñ̤Ѵ) */
#define radtodeg(rad)   (fmod((rad),2*PI)*180.0/PI)

/* ɥ礭ѿ (¤ϺʤΤ) */
static unsigned int winwidth, winheight;
/* space() ǻꤷκɸ (plot(5) κɸ) */
static int spx0, spy0;
static int denominator, numerator;      /* window / space          */
/* κɸ (plot(5)κɸ) */
static int cpx, cpy;

typedef struct { int x1,y1,x2,y2; } rectangle;
static rectangle window_rec;

/* */

static int do_record = 1;

#define low(c)  ((c) & 0x00ff)
#define high(c) (((c) & 0xffff) >> 8)

#ifdef  HIGHBYTEFIRST
#define DEFAULT_FIRST_BYTE_IS_HIGH      1
#else
#define DEFAULT_FIRST_BYTE_IS_HIGH      0
#endif
int firstbyte_is_high = DEFAULT_FIRST_BYTE_IS_HIGH;

/* ե short Υǡ񤭹 */
static void writeshort(int s, FILE *f)
{
    if (firstbyte_is_high) {
        fputc(high(s), f); fputc(low(s), f);
    }
    else {
        fputc(low(s), f); fputc(high(s), f);
    }
}

/* ߵƤ޷ǡե˽񤭹 */
void mkplot(char *fname)
{
    FILE *wf;
    if ((wf = fopen(fname, "w")) == NULL) {
            fprintf(stderr, "cannot open %s\n", fname);
    }
    else {
        int i, c, num_of_commands;
        /* space() */
        fputc('s',wf);
        writeshort(window_rec.x1, wf); writeshort(window_rec.y1, wf);
        writeshort(window_rec.x2, wf); writeshort(window_rec.y2, wf);
        /* write rec[] */
	num_of_commands = startreading();
        for (i = 0; i < num_of_commands;) {
            fputc(c = readdata(), wf);
            switch (c) {
            case 'm':
            case 'n':
            case 'p':
                writeshort(readdata(), wf); writeshort(readdata(), wf);
		i += 3;
                break;
            case 'l':
                writeshort(readdata(), wf); writeshort(readdata(), wf);
                writeshort(readdata(), wf); writeshort(readdata(), wf);
		i += 5;
                break;
            case 'c':
                writeshort(readdata(), wf); writeshort(readdata(), wf);
                writeshort(readdata(), wf);
		i += 4;
                break;
            case 'a':
                writeshort(readdata(), wf); writeshort(readdata(), wf);
                writeshort(readdata(), wf); writeshort(readdata(), wf);
                writeshort(readdata(), wf); writeshort(readdata(), wf);
		i += 7;
                break;
            case 't':
            case 'f':
		/* string of plot(3) must not contain NEWLINE
		 * string of plot(5) must terminate with NEWLINE */
                while ((c = readdata()) != 0) { fputc(c, wf); i++; }
		i += 2;
		fputc('\n', wf);
                break;
            default:
                fprintf(stderr, "mkplot: unkown command %c=0x%x\n", c, c);
		break;
            }
        }
        fclose(wf);
	finishreading();
    }
}

void mkplot2(char *basename, int n)
{
    char fname[1024];
    if (n < 100000)
	sprintf(fname, "%s%05d.plot", basename, n);
    else
	sprintf(fname, "%s%d.plot", basename, n);
    mkplot(fname);
}

/* for Fortran */
void mkplot_(char *fname) { toasciz(fname); mkplot(fname); }

void mkplot2_(char *basename, int *n)
{ deltailsp(basename); mkplot2(basename, *n); }

/* ǥեȤΥɥΥ */
#define X_WIDTH 150.0
#define Y_WIDTH 150.0

/* ߤΥɥΥ (ɥåȿñ) Ĵ٤ */
static void current_window_size(unsigned int *wx, unsigned int *wy)
{
  /* ΤȤ */
  *wx = X_WIDTH; *wy = Y_WIDTH;
}

/* plot 饤֥ꥣν */
void openpl(void)
{
  /* GLSC ν */
  g_init("META", (G_REAL)X_WIDTH, (G_REAL)Y_WIDTH);
  g_device(G_BOTH);

  /*  */
  g_def_line(0, G_BLACK, 0, G_LINE_SOLID);
  g_def_line(1, G_BLACK, 0, G_LINE_DOTS);
  /* ʸ° */
  g_def_text(0, G_BLACK, 1);
  /* Τ򤹤 */
  g_sel_line(0);
  g_sel_text(0);

  /* ߤΥɥΥ (Windows 9x β̤ñ̤) */
  current_window_size(&winwidth, &winheight);
}

void check_expose(void) { }

/* for Fortran */
void check_expose_(void)
{
    check_expose();
}

void chkexp_(void)
{
    check_expose();
}

/* plot 饤֥ꥣνλ */
void closepl(void)
{
    g_sleep(-1.0);
    g_term();
}

void space(short int x1, short int y1, short int x2, short int y2)
{
  int dx, dy;

  window_rec.x1 = x1; window_rec.y1 = y1;
  window_rec.x2 = x2; window_rec.y2 = y2;

  dx = x2 - x1; dy = y2 - y1;
  if (dx * winheight > dy * winwidth) {
    numerator = winwidth;
    denominator = dx;
    spx0 = x1;
    spy0 = y1 - (dx * winheight / winwidth - dy) / 2;
  }
  else {
    numerator = winheight;
    denominator = dy;
    spx0 = x1 - (dy * winwidth / winheight - dx) / 2;
    spy0 = y1;
  }

  g_def_scale(0,
	      (G_REAL)x1, (G_REAL)x2, (G_REAL)y1, (G_REAL)y2,
	      0.0, 0.0, X_WIDTH, Y_WIDTH);
  g_sel_scale(0);
}

/* plot(5) κɸ饦ɥκɸ(ɥåñ)ؤѴ --- x */
static int winx(int x)
{ return (x  - spx0) * numerator / denominator; }

/* plot(5) κɸ饦ɥκɸ(ɥåñ)ؤѴ --- y */
static int winy(int y)
{ return winheight - ((y  - spy0) * numerator / denominator); }

/* ɥκɸ(ɥåñ) plot(5) κɸؤѴ --- x */
static int plx(int wx)
{ return spx0 + wx * denominator / numerator; }

/* ɥκɸ(ɥåñ) plot(5) κɸؤѴ --- y */
static int ply(int wy)
{ return spy0 + (winheight - wy) * denominator / numerator; }

/* 褷޷ξõ */
void erase(void)
{
    g_cls();
}

/*  */
void redraw(void)
{
    int i, do_record_save, num_of_commands;

    /* check the current window size */
    current_window_size(&winwidth, &winheight);
    space(window_rec.x1, window_rec.y1, window_rec.x2, window_rec.y2);
    /* not to record commmads to redraw the window */
    do_record_save = do_record;
    do_record = 0;
    /* clear the window */
    g_cls();
    /* do redrawing */
    num_of_commands = startreading();
    for (i = 0; i < num_of_commands;) {
        int c, x, y, x1, y1, x2, y2, r;
        char *s, str[512];
        switch (c = readdata()) {
            case 'm': x = readdata(); y = readdata(); move(x,y); i += 3; break;
            case 'n': x = readdata(); y = readdata(); cont(x,y); i += 3; break;
            case 'p': x = readdata(); y = readdata(); point(x,y); i += 3; break;
            case 'l': x1 = readdata(); y1 = readdata();
                      x2 = readdata(); y2 = readdata();
                      line(x1,y1,x2,y2); i += 5; break;
            case 'c': x = readdata(); y = readdata(); r = readdata();
                      circle(x,y,r); i += 4; break;
            case 'a': x = readdata(); y = readdata();
                      x1 = readdata(); y1 = readdata();
                      x2 = readdata(); y2 = readdata();
                      arc(x,y,x1,y1,x2,y2); i += 7; break;
            case 't': for (s = str; (*s++ = readdata()); ); label(str);
		      i += (strlen(str) + 2); break;
            case 'f': for (s = str; (*s++ = readdata()); ); linemod(str); 
		      i += (strlen(str) + 2); break;
            default:
                fprintf(stderr, "unknown command %c 0x%02x\n", c, c);
		goto QUIT;
                break;
        }
    }
QUIT:
    finishreading();
    do_record = do_record_save;
}

/* for Fortran */
void redraw_(void)
{
    redraw();
}

static int last_record_point_x = 0;
static int last_record_point_y = 0;

void move(short int x, short int y)
{
    cpx = x; cpy = y;
    g_move((G_REAL)x, (G_REAL)y);
    /* to redraw */
    if (do_record) {
	    record('m'); record(x); record(y);
	    last_record_point_x = x; last_record_point_y = y;
    }
}

void cont(short int x, short int y)
{
    g_plot((G_REAL)x, (G_REAL)y);
    /* current point */
    cpx = x; cpy = y;
    /* to redraw */
    if (do_record) { record('n'); record(x); record(y); }
}

void line(short int x1, short int y1, short int x2, short int y2)
{
    g_move((G_REAL)x1, (G_REAL)y1);
    g_plot((G_REAL)x2, (G_REAL)y2);
    /* current point */
    cpx = x2; cpy = y2;
    /* to redraw */
    if (do_record) {
        record('l'); record(x1); record(y1); record(x2); record(y2);
    }
}

void point(short int x, short int y)
{
    g_point((G_REAL)x, (G_REAL)y);
    cpx = x; cpy = y;
    /* to redraw */
    if (do_record) {
	record('p'); record(x); record(y);
	last_record_point_x = x; last_record_point_y = y;
    }
}

void circle(short int x, short int y, short int r)
{
    int d; /* diameter */

    d = 2 * r * numerator / denominator;
 /*  XDrawArc(dpy, win, gc, winx(x - r), winy(y + r), d, d, 0, 360 * 64); */
    cpx = x; cpy = y;
    /* to redraw */
    if (do_record) {
	record('c'); record(x); record(y); record(r);
	last_record_point_x = x; last_record_point_y = y;
    }
}

void arc(short int x, short int y,
         short int x1, short int y1, short int x2, short int y2)
{
    int d, r;
    int ang1, ang2, angle;

    r = rint(hypot((G_REAL)(x - x1), (G_REAL)(y - y1)));
    d = 2 * r * numerator / denominator;

    ang1 = rint(radtodeg(atan2((G_REAL)(y1 - y), (G_REAL)(x1 - x))));
    ang2 = rint(radtodeg(atan2((G_REAL)(y2 - y), (G_REAL)(x2 - x))));
    angle = ang2 - ang1;
    if (angle < 0) angle += 360;

/*
    XDrawArc(dpy, win, gc,
             winx(x - r), winy(y + r), d, d, ang1*64, angle*64);
*/

    cpx = x; cpy = y;
    /* to redraw */
    if (do_record) {
        record('a');
        record(x); record(y); record(x1); record(y1); record(x2); record(y2);
	last_record_point_x = x; last_record_point_y = y;
    }
}

void label(char *s)
{
    int len;
    char buf[512];
    /* delete newline characters at the tail. -- not necessary */
    strncpy(buf, s, sizeof(buf));
    deltailnewline(buf);
    deltailsp(buf);
    /* */
    if ((len = strlen(buf)) > 0) {
        g_text(winx(cpx), winy(cpy), buf);
        /* to redraw */
        if (do_record) {
            int i;
            record('t');
            for (i = 0; i < len; i++) record(buf[i]);
            record(0);
        }
    }
}

static char name[][12]
 = {"solid", "dotted", "dotdashed", "shortdashed", "longdashed"};

void linemod(char *s)
{
    int type;

    for (type = 0; type < 5; type++)
        if (strcmp(s, name[type]) == 0) {
            if (type == 0)
		g_sel_line(0);
            else
		g_sel_line(1);
            /* to redraw */
            if (do_record) {
                int c;
                record('f'); while ((c = *s++)) record(c); record(0);
            }
            return; /* exit this function */
        }
    fprintf(stderr, "linemod: unknown mode \"%s\"\n", s);
    return;
}

/*************************************************************************
 *                      useful for animation
 *************************************************************************/

static void warning(char *msg)
{
  fprintf(stderr, "%s() is not supported\n", msg);
}

void xor() { warning("xor"); }
void set() { warning("set"); }
void xflush() { warning("xflush"); }
void xsync() { }
void xor_() { xor(); }
void set_() { set(); }
void xflush_() { xflush(); }
void xsync_() { xsync(); }

/*************************************************************************
 *                              mouse
 *************************************************************************/

void mouse(int *l, int *m, int *r, int *x, int *y)
{
    *l = *m = *r = 0;
}

static void deltailnewline(char *s)
{
    char *tailpos;
    if (*s == 0) return;
    if (*s == '\n') {
        /* Ƭ '\n' ʤСʤޤꤨʤȤǤɡ
        ʸˤ롣 */
        *s = 0; return;
    }
    /* ǤƬʸ '\n' ǤʤĹäʸ */
    /* tailpos  0 ǤʤǸʸΰ */
    tailpos = s + (strlen(s) - 1);
    while (*tailpos == '\n') tailpos--;
    *(tailpos+1) = 0;
}

static void deltailsp(char *s)
{
    int len, i;
    len = strlen(s);
    for (i = len - 1; i >= 0; i--)
	if (s[i] != ' ')
	    return;
	else
	    s[i] = '\0';
}

/* */
static int toasciz(char *s)
{
    int i;
    for (i = 0; i < 200; i++)
	if (s[i] == 0)
		return 0;
	else if (s[i] == ' ') {
	    s[i] = 0;
	    return 0;
	}
    return 1;
}
