/*
 * xplot.c
 *      written by Fujio
 *      modified by Masashi Katsurada
 */

#define USE_BACKINGSTORE
/* #define USE_MY_LINE */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <X11/Xlib.h>

#include "MyBuffer.h"
#include "fplot.h"

static void current_window_size(unsigned int *, unsigned int *);
static void writeshort(int, FILE *);
static int winx(int);
static int winy(int);
static int plx(int);
static int ply(int);
static int comp_flag(int, int);
static void sublinew(int *, int *, int *, int *, int *);
static void clip_line(int, int, int, int);
static void no_clip_line(int, int, int, int);
static void no_clip_point(int, int);
static void clip_point(int, int);
static int toasciz(char *);
static void deltailnewline(char *);
static void deltailsp(char *);

#ifndef M_PI
#define PI      3.14159265358979323846
#else
#define PI      M_PI
#endif

#define radtodeg(rad)   (fmod((rad),2*PI)*180.0/PI)

#define X_AXIS  500
#define Y_AXIS  500
static Display *dpy;
static Window root, win;
static GC gc;
static int screen, black, white;
static XEvent event;
static XFontStruct *font;
static char *fontname = "fixed";
static int FONTcomeyet = 0;

static unsigned int winwidth, winheight;/* the size of the window  */
static int spx0, spy0;                  /* space function          */
static int denominator, numerator;      /* window / space          */
static int cpx, cpy;                    /* current graphical point */

typedef struct { int x1,y1,x2,y2; } rectangle;
static rectangle window_rec;
static rectangle clip_win;

static void current_window_size(unsigned int *wx, unsigned int *wy)
{
    int x, y;
    unsigned int bw, depth;    /* dummy */
    XGetGeometry(dpy, win, &root, &x, &y, wx, wy, &bw, &depth);
}

/* */

static int do_record = 1;

#define low(c)  ((c) & 0x00ff)
#define high(c) (((c) & 0xffff) >> 8)

#ifdef  HIGHBYTEFIRST
#define DEFAULT_FIRST_BYTE_IS_HIGH      1
#else
#define DEFAULT_FIRST_BYTE_IS_HIGH      0
#endif
int firstbyte_is_high = DEFAULT_FIRST_BYTE_IS_HIGH;

static void writeshort(int s, FILE *f)
{
    if (firstbyte_is_high) {
        fputc(high(s), f); fputc(low(s), f);
    }
    else {
        fputc(low(s), f); fputc(high(s), f);
    }
}

void mkplot(char *fname)
{
    FILE *wf;
    if ((wf = fopen(fname, "w")) == NULL) {
            fprintf(stderr, "cannot open %s\n", fname);
    }
    else {
        int i, c, num_of_commands;
        /* space() */
        fputc('s',wf);
        writeshort(window_rec.x1, wf); writeshort(window_rec.y1, wf);
        writeshort(window_rec.x2, wf); writeshort(window_rec.y2, wf);
        /* write rec[] */
	num_of_commands = startreading();
        for (i = 0; i < num_of_commands;) {
            fputc(c = readdata(), wf);
            switch (c) {
            case 'm':
            case 'n':
            case 'p':
                writeshort(readdata(), wf); writeshort(readdata(), wf);
		i += 3;
                break;
            case 'l':
                writeshort(readdata(), wf); writeshort(readdata(), wf);
                writeshort(readdata(), wf); writeshort(readdata(), wf);
		i += 5;
                break;
            case 'c':
                writeshort(readdata(), wf); writeshort(readdata(), wf);
                writeshort(readdata(), wf);
		i += 4;
                break;
            case 'a':
                writeshort(readdata(), wf); writeshort(readdata(), wf);
                writeshort(readdata(), wf); writeshort(readdata(), wf);
                writeshort(readdata(), wf); writeshort(readdata(), wf);
		i += 7;
                break;
            case 't':
            case 'f':
		/* string of plot(3) must not contain NEWLINE
		 * string of plot(5) must terminate with NEWLINE */
                while ((c = readdata()) != 0) { fputc(c, wf); i++; }
		i += 2;
		fputc('\n', wf);
                break;
            default:
                fprintf(stderr, "mkplot: unkown command %c=0x%x\n", c, c);
		break;
            }
        }
        fclose(wf);
	finishreading();
    }
}

void mkplot2(char *basename, int n)
{
    char fname[1024];
    if (n < 100000)
	sprintf(fname, "%s%05d.plot", basename, n);
    else
	sprintf(fname, "%s%d.plot", basename, n);
    mkplot(fname);
}

/* for Fortran */
void mkplot_(char *fname, int n)
{
  char *s;
  s = malloc(n+1);
  strncpy(s,fname,n);
  s[n] = 0;
  mkplot(s);
  free(s);
}

void mkplot2_(char *basename, int *n)
{ deltailsp(basename); mkplot2(basename, *n); }

void openpl(void)
{

    if ((dpy = XOpenDisplay("")) == NULL) { /* connect to X server    */
        fprintf(stderr, "cannot open display %s\n", getenv("DISPLAY"));
        exit(1);
    }

    root = XDefaultRootWindow(dpy); 
    screen = XDefaultScreen(dpy);
    black = XBlackPixel(dpy, screen);
    white = XWhitePixel(dpy, screen);

    /* Open window for INPUT/OUTPUT(This is normal) */
    win = XCreateSimpleWindow(dpy, root, 0, 0, X_AXIS, Y_AXIS, 1, black, white);

    XSelectInput(dpy, win,
                 ExposureMask | ButtonPressMask | ButtonReleaseMask |
                 Button1MotionMask);

    gc = XCreateGC(dpy, win, 0, NULL);  /* use graphic */
    XSetForeground(dpy, gc, black);

    XStoreName(dpy, win, "fplot");

    XMapWindow(dpy, win);
    XFlush(dpy);

    XNextEvent(dpy, &event);    /* wait for the window open in real screen */

    current_window_size(&winwidth, &winheight);

#ifdef USE_BACKINGSTORE
    {
      XSetWindowAttributes att;
      att.backing_store = WhenMapped;
      XChangeWindowAttributes(dpy, win, CWBackingStore, &att);
    }
#endif

    XSelectInput(dpy, win, ExposureMask);
    XFlush(dpy);
}

#define	USENEW

void check_expose(void)
{
#ifdef	USENEW
    if (XCheckTypedWindowEvent(dpy, win, Expose, &event)) {
	redraw();
    }
#else
    while (XEventsQueued(dpy, QueuedAfterReading)) {
        XNextEvent(dpy, &event);
        switch(event.type) {
            case Expose:
                if (event.xexpose.count == 0)
		    while (XEventsQueued(dpy, QueuedAfterReading) > 0) {
			XEvent ahead;
			XPeekEvent(dpy, &ahead);
			if (ahead.type == Expose) XNextEvent(dpy, &event);
		    }
		    redraw();
                break;
            default:
                break;
        }
    }
#endif
}

/* for Fortran */
void check_expose_(void)
{
    check_expose();
}

void chkexp_(void)
{
    check_expose();
}

void closepl(void)
{
    XSelectInput(dpy, win, ButtonPressMask | ExposureMask);
    while (1) {
        XNextEvent(dpy, &event);/* wait button press before break the window */
        switch(event.type) {
            case ButtonPress:
                if (FONTcomeyet) XFreeFont(dpy, font);
                XFreeGC(dpy, gc);
                XDestroyWindow(dpy, win);
                XCloseDisplay(dpy);
                return;
            case Expose:
                if (event.xexpose.count == 0)
#ifdef	OLD
		    redraw();
#else
		    while (XEventsQueued(dpy, QueuedAfterReading) > 0) {
			XEvent ahead;
			XPeekEvent(event.xexpose.display, &ahead);
			if (ahead.type != Expose) break;
			if (ahead.xexpose.window != event.xexpose.window) break;
			XNextEvent(dpy, &event);
		    }
		    redraw();
#endif
                break;
            default:
                break;
        }
    }
}

void space(short int x1, short int y1, short int x2, short int y2)
{
    int dx, dy;

    /* record the arguments */
    window_rec.x1 = x1; window_rec.y1 = y1;
    window_rec.x2 = x2; window_rec.y2 = y2;
    clip_win.x1 = x1; clip_win.y1 = y1;
    clip_win.x2 = x2; clip_win.y2 = y2;

    /* */
    dx = x2 - x1;
    dy = y2 - y1;
    if (dx <= 0 || dy <= 0) {
        fprintf(stderr, "space parameter error %d %d %d %d\n", x1, y1, x2, y2);
        return;
    }
    if (dx * winheight > dy * winwidth) {
        numerator = winwidth;
        denominator = dx;
        spx0 = x1;
        spy0 = y1 - (dx * winheight / winwidth - dy) / 2;
    }
    else {
        numerator = winheight;
        denominator = dy;
        spx0 = x1 - (dy * winwidth / winheight - dx) / 2;
        spy0 = y1;
    }
}

static int winx(int x)
{ return (x  - spx0) * numerator / denominator; }

static int winy(int y)
{ return winheight - ((y  - spy0) * numerator / denominator); }

static int plx(int wx)
{ return spx0 + wx * denominator / numerator; }

static int ply(int wy)
{ return spy0 + (winheight - wy) * denominator / numerator; }

void erase(void)
{
    clear_queue();
    XClearWindow(dpy, win);
}

void redraw(void)
{
    int i, do_record_save, num_of_commands;

    /* check the current window size */
    current_window_size(&winwidth, &winheight);
    space(window_rec.x1, window_rec.y1, window_rec.x2, window_rec.y2);
    /* not to record commmads to redraw the window */
    do_record_save = do_record;
    do_record = 0;
    /* clear the window */
    XClearWindow(dpy, win);
    /* do redrawing */
    num_of_commands = startreading();
    for (i = 0; i < num_of_commands;) {
        int c, x, y, x1, y1, x2, y2, r;
        char *s, str[512];
        switch (c = readdata()) {
            case 'm': x = readdata(); y = readdata(); move(x,y); i += 3; break;
            case 'n': x = readdata(); y = readdata(); cont(x,y); i += 3; break;
            case 'p': x = readdata(); y = readdata(); point(x,y); i += 3; break;
            case 'l': x1 = readdata(); y1 = readdata();
                      x2 = readdata(); y2 = readdata();
                      line(x1,y1,x2,y2); i += 5; break;
            case 'c': x = readdata(); y = readdata(); r = readdata();
                      circle(x,y,r); i += 4; break;
            case 'a': x = readdata(); y = readdata();
                      x1 = readdata(); y1 = readdata();
                      x2 = readdata(); y2 = readdata();
                      arc(x,y,x1,y1,x2,y2); i += 7; break;
            case 't': for (s = str; (*s++ = readdata()); ); label(str);
		      i += (strlen(str) + 2); break;
            case 'f': for (s = str; (*s++ = readdata()); ); linemod(str); 
		      i += (strlen(str) + 2); break;
            default:
                fprintf(stderr, "unknown command %c 0x%02x\n", c, c);
		goto QUIT;
                break;
        }
    }
QUIT:
    finishreading();
    do_record = do_record_save;
}

/* for Fortran */
void redraw_(void)
{
    redraw();
}

/* ****************************************************************
 *                           clipping
 * **************************************************************** */

static int comp_flag(int x, int y)
{
    int flag;
    flag = 0;		
    if (y > clip_win.y2) flag = 1; else if (y < clip_win.y1) flag = 2;
    if (x > clip_win.x2) flag += 4; else if (x < clip_win.x1) flag += 8;
    return flag;
}

static void sublinew(int *x1, int *y1, int *x2, int *y2, int *flag)
{
    if (*flag & 1)  {
        *x2 = *x1 + (clip_win.y2 - *y1) * (*x2 - *x1) / (*y2 - *y1);
        *y2 = clip_win.y2;
        *flag = comp_flag(*x2,*y2);
    }
    else if (*flag & 2) {
        *x2 = *x1 + (clip_win.y1 - *y1) * (*x2 - *x1) / (*y2 - *y1);
        *y2 = clip_win.y1;
        *flag = comp_flag(*x2,*y2);
    }
    if (*flag & 4)  {
        *y2 = *y1 + (clip_win.x2 - *x1) * (*y2 - *y1) / (*x2 - *x1);
        *x2 = clip_win.x2;
    }
    else if (*flag & 8)  {
        *y2 = *y1 + (clip_win.x1 - *x1) * (*y2 - *y1) / (*x2 - *x1);
        *x2 = clip_win.x1;
    }
}

static void clip_line(int x1, int y1, int x2, int y2)
{
    /* main part of ClippingLine */
    int flag1, flag2;
    flag1 = comp_flag(x1, y1);
    flag2 = comp_flag(x2, y2);
    if (flag1 & flag2) return;
    if (flag1 == 0) {
        if (flag2) sublinew(&x1, &y1, &x2, &y2, &flag2);
        no_clip_line(x1, y1, x2, y2);
    }
    else if (flag2 == 0) {
        sublinew(&x2, &y2, &x1, &y1, &flag1);
        no_clip_line(x1, y1, x2, y2);
    }
    else {
        sublinew(&x2, &y2, &x1, &y1, &flag1);
        flag1 = comp_flag(x1, y1);
        if (flag1 == 0) {
            sublinew(&x1, &y1, &x2, &y2, &flag2);
            no_clip_line(x1, y1, x2, y2);
        }
    }
}
/* ****************************************************************
 *                        clipping end
 * **************************************************************** */

#define	RECORD_IN_NO_CLIP_LINE
#define	RECORD_IN_NO_CLIP_POINT

static int last_record_point_x = 0;
static int last_record_point_y = 0;

static void no_clip_line(int x1, int y1, int x2, int y2)
{
    XDrawLine(dpy, win, gc, winx(x1), winy(y1), winx(x2), winy(y2));
#ifdef	RECORD_IN_NO_CLIP_LINE
    /* to redraw */
    if (do_record) {
	if (last_record_point_x != x1 || last_record_point_y != y1) {
	    record('m'); record(x1); record(y1);
	}
	record('n'); record(x2); record(y2);
	last_record_point_x = x2; last_record_point_y = y2;
    }
#endif
}

#define CLIP
#ifdef  CLIP
#define LINE(x1,y1,x2,y2)       clip_line(x1,y1,x2,y2)
#define POINT(x,y)              clip_point(x,y)
#else
#define LINE(x1,y1,x2,y2)       no_clip_line(x1,y1,x2,y2)
#define POINT(x,y)              no_clip_point(x,y)
#endif

void move(short int x, short int y)
{
    cpx = x; cpy = y;
    /* to redraw */
    if (do_record) {
	if (last_record_point_x != x || last_record_point_y != y) {
	    record('m'); record(x); record(y);
	    last_record_point_x = x; last_record_point_y = y;
	}
    }
}

void cont(short int x, short int y)
{
    LINE(cpx, cpy, x, y);
    cpx = x; cpy = y;
#ifndef	RECORD_IN_NO_CLIP_LINE
    /* to redraw */
    if (do_record) { record('n'); record(x); record(y); }
#endif
}

void line(short int x1, short int y1, short int x2, short int y2)
{
#ifdef	OLDERCODE
    LINE(x1, y1, x2, y2);
    cpx = x2; cpy = y2;
    /* to redraw */
    if (do_record) {
#ifndef	RECORD_IN_NO_CLIP_LINE
        record('l'); record(x1); record(y1); record(x2); record(y2);
#endif
    }
#else
    move(x1, y1); cont(x2, y2);
#endif
}

static void no_clip_point(int x, int y)
{
    XDrawPoint(dpy, win, gc, winx(x), winy(y));
#ifdef RECORD_IN_NO_CLIP_POINT
    /* to redraw */
    if (do_record) {
	record('p'); record(x); record(y);
	last_record_point_x = x; last_record_point_y = y;
    }
#endif
}

static void clip_point(int x, int y)
{
    if (clip_win.x1 <= x && x <= clip_win.x2 &&
        clip_win.y1 <= y && y <= clip_win.y2)
        no_clip_point(x, y);
}

void point(short int x, short int y)
{
    POINT(x,y);
    cpx = x; cpy = y;
#ifndef RECORD_IN_NO_CLIP_POINT
    /* to redraw */
    if (do_record) {
	record('p'); record(x); record(y);
	last_record_point_x = x; last_record_point_y = y;
    }
#endif
}

void circle(short int x, short int y, short int r)
{
    int d; /* diameter */

    d = 2 * r * numerator / denominator;
    XDrawArc(dpy, win, gc, winx(x - r), winy(y + r), d, d, 0, 360 * 64);
    cpx = x; cpy = y;
    /* to redraw */
    if (do_record) {
	record('c'); record(x); record(y); record(r);
	last_record_point_x = x; last_record_point_y = y;
    }
}

void arc(short int x, short int y,
         short int x1, short int y1, short int x2, short int y2)
{
    int d, r;
    int ang1, ang2, angle;

    r = rint(hypot((double)(x - x1), (double)(y - y1)));
    d = 2 * r * numerator / denominator;

    ang1 = rint(radtodeg(atan2((double)(y1 - y), (double)(x1 - x))));
    ang2 = rint(radtodeg(atan2((double)(y2 - y), (double)(x2 - x))));
    angle = ang2 - ang1;
    if (angle < 0) angle += 360;

    XDrawArc(dpy, win, gc,
             winx(x - r), winy(y + r), d, d, ang1*64, angle*64);
    cpx = x; cpy = y;
    /* to redraw */
    if (do_record) {
        record('a');
        record(x); record(y); record(x1); record(y1); record(x2); record(y2);
	last_record_point_x = x; last_record_point_y = y;
    }
}

void label(char *s)
{
    int len;
    char buf[512];
    /* I do not know how to do clipping! */
    /* delete newline characters at the tail. -- not necessary */
    strncpy(buf, s, sizeof(buf));
    deltailnewline(buf);
    deltailsp(buf);
    /* */
    if ((len = strlen(buf)) > 0) {
        /* FONTcomeyet indicates whether Font is active or not. */
        if (FONTcomeyet == 0) {       
            FONTcomeyet = 1;    /* Font is on use */
            font = XLoadQueryFont(dpy, fontname);
            XSetFont(dpy, gc, font->fid);
        }
        /* draw a string */
        XDrawString(dpy, win, gc, winx(cpx), winy(cpy), buf, len);
        /* to redraw */
        if (do_record) {
            int i;
            record('t');
            for (i = 0; i < len; i++) record(buf[i]);
            record(0);
        }
    }
}

static char name[][12]
 = {"solid", "dotted", "dotdashed", "shortdashed", "longdashed"};

static int lenname[] = {5, 6, 9, 11, 10};

void linemod(char *s)
{
    int type;
#ifdef USE_MY_LINE
    static char line[5][4]={{10}, {5, 5}, {10, 5}, {20, 5}, {20, 5, 5, 5}};
#endif

    for (type = 0; type < 5; type++)
        if (strncmp(s, name[type], lenname[type]) == 0) {
#ifdef USE_MY_LINE
            XSetDashes(dpy, gc, 0, line[type], num);
#else
            if (type == 0)
                XSetLineAttributes(dpy, gc, 1, LineSolid, CapButt, JoinMiter);
            else
                XSetLineAttributes(dpy, gc, 1, LineOnOffDash, CapButt, JoinMiter);
#endif
            /* to redraw */
            if (do_record) {
                int c;
                record('f'); while ((c = *s++)) record(c); record(0);
            }
            return; /* exit this function */
        }
    fprintf(stderr, "linemod: unknown mode %s\n", s);
    return;
}

/*************************************************************************
 *                      useful for animation
 *************************************************************************/

void xor() { XSetFunction(dpy, gc, GXxor); }
void set() { XSetFunction(dpy, gc, GXcopy); }
void xflush() { XFlush(dpy); }
void xsync() { XSync(dpy, False); }
void xor_() { xor(); }
void set_() { set(); }
void xflush_() { xflush(); }
void xsync_() { xsync(); }

/*************************************************************************
 *                              mouse
 *************************************************************************/

void mouse(int *l, int *m, int *r, int *x, int *y)
{
    *l = *m = *r = 0;
    XSelectInput(dpy, win, ExposureMask | ButtonPressMask);
    while (1) {
        XNextEvent(dpy, &event);
        switch (event.type) {
            case ButtonPress:
                if (event.xbutton.button == 1)
                    *l = 1;
                else if (event.xbutton.button == 2)
                    *m = 1;
                else if (event.xbutton.button == 3)
                    *r = 1;
                *x = plx(event.xbutton.x);
                *y = ply(event.xbutton.y);
                goto QUIT_LOOP;
            case Expose:
                if (event.xexpose.count == 0) redraw();
                break;
            default:
                break;
        }
    }
QUIT_LOOP:
    XSelectInput(dpy, win, ExposureMask);
}

static void deltailnewline(char *s)
{
    char *tailpos;
    if (*s == 0) return;
    if (*s == '\n') {
        /* Ƭ '\n' ʤСʤޤꤨʤȤǤɡ
        ʸˤ롣 */
        *s = 0; return;
    }
    /* ǤƬʸ '\n' ǤʤĹäʸ */
    /* tailpos  0 ǤʤǸʸΰ */
    tailpos = s + (strlen(s) - 1);
    while (*tailpos == '\n') tailpos--;
    *(tailpos+1) = 0;
}

static void deltailsp(char *s)
{
    int len, i;
    len = strlen(s);
    for (i = len - 1; i >= 0; i--)
	if (s[i] != ' ')
	    return;
	else
	    s[i] = '\0';
}

/* */
static int toasciz(char *s)
{
    int i;
    for (i = 0; i < 200; i++)
	if (s[i] == 0)
		return 0;
	else if (s[i] == ' ') {
	    s[i] = 0;
	    return 0;
	}
    return 1;
}
