* implicit4.f -- 陰的スキームで熱方程式を解く
*   「微分方程式と計算機演習」第11章 p237 のプログラムを少し修正したもの
C     ** IMPLICIT FINITE DIFFERENCE METHOD **
C     **         FOR HEAT EQUATION         **
C     nfunc : 関数の番号（2種類の関数を用意している）
C     theta : スキームのパラメーター
C     N : 分割の数
C     t : 時刻
C     tau : 時間の増分
C     Tmax : 時刻の上限
C     h : 空間のメッシュ幅
C     a,b,c : 係数行列
C     u : DIMENSION FOR UNKNOWN FUNCTION
      program iheat
      integer ndim,N,nfunc,i,nstep
      real theta,h,Tmax,tau,c1,c2,c3,c4,lambda,f,t
      parameter (ndim=100)
      real a(ndim-1),b(ndim-1),c(ndim-1),ff(ndim-1),u(0:ndim)
*     入力
      write(*,*) '入力して下さい : nfunc='
      read(*,*) nfunc
      write(*,*) '入力して下さい : θ='
      read(*,*) theta
      write(*,*) '入力して下さい : N='
      read(*,*) N
      if (N .le. 1) stop
*      write(*,*) '入力して下さい : τ='
*      read(*,*) tau
      write(*,*) '入力して下さい : λ='
      read(*,*) lambda
*
      h=1.0/float(N)
      tau=lambda*h*h
      write(*,*) ' 時間の増分τ=', tau, 'になりました。'
*      lambda=tau/h**2
*
      write(*,*) '入力して下さい : Tmax='
      read(*,*) Tmax
*
c     係数行列の計算
      c1=-theta*lambda
      c2=1.0+2.0*theta*lambda
      c3=1.0-2.0*(1.0-theta)*lambda
      c4=(1.0-theta)*lambda
      do 1 i=1,N-1
	a(i)=c1
	b(i)=c2
	c(i)=c1
    1 continue
c     初期値
      do 2 i=0,N
	u(i)=f(i*h,nfunc)
    2 continue
c      グラフィックスの準備
      call openpl
      call fspace2(-0.05,-0.05,1.05,1.05)
      call fline(-0.05,0.0,1.05,0.0)
      call fline( 0.0,-0.05,0.0,1.05)
c      t=0における解（＝初期条件）の様子を描く
      call fmove(0.0,u(0))
      do i=0,N
	call fcont(i*h,u(i))
      end do
      call xflush
c      出力（t=0）
      write(*,100) 0.0,(i,u(i),i=0,N)
  100 format(1H ,'T=',1PE12.4/
     1       1H ,5(2X,'I',5X,'u(i)',4X)/
     2      (1H ,5(i3,1PE12.4,1X)))
      nstep=0
c     計算 --- entrance of the loop
    3 continue
      do 4 i=1,N-1
    4 ff(i)=c3*u(i)+c4*(u(i-1)+u(i+1))
      call trid2(a,b,c,ff,N-1,nstep)
      do 5 i=1,N-1
    5 u(i)=ff(i)
      u(0)=0.0
      u(N)=0.0
      nstep=nstep+1
      t=tau*float(nstep)
c      この時刻における解の様子を描く
      call erase
      call fmove(0.0,u(0))
      do i=1,N
	call fcont(i*h,u(i))
      end do
      call xflush
c     出力
*      write(*,100) t,(i,u(i),i=0,N)
      if (t+tau.lt.Tmax+0.1*tau) goto 3
c
      call mkplot("heat.plot")
c      グラフィックスの後始末
      call closepl
c
      end
**********************************************************
**********************************************************
*      三項方程式を解く
      subroutine trid(a,b,c,f,n,id)
      integer n,id
      real a(*),b(*),c(*),f(*)
      integer i
      real aa
c     前進消去(forward elimination)
      do 1 i=1,n-1
	aa=a(i+1)/b(i)
	if (id .eq. 0) b(i+1)=b(i+1)-aa*c(i)
	f(i+1)=f(i+1)-aa*f(i)
    1 continue
c     後退代入(backward substitution)
      f(n)=f(n)/b(n)
      do 2 i=n-1,1,-1
    2 f(i)=(f(i)-c(i)*f(i+1))/b(i)
      end
**********************************************************
      subroutine trid2(a,b,c,f,n,id)
      integer n,id
      real a(*),b(*),c(*),f(*)
      integer i
c     前進消去(forward elimination)
      if (id .eq. 0) then
	do 1 i=1,n-1
	  b(i+1)=b(i+1)-c(i)*a(i+1)/b(i)
    1   continue
      endif
      do 2 i=1,n-1
	f(i+1)=f(i+1)-f(i)*a(i+1)/b(i)
    2 continue
c     後退代入(backward substitution)
      f(n)=f(n)/b(n)
      do 3 i=n-1,1,-1
    3 f(i)=(f(i)-c(i)*f(i+1))/b(i)
      end
**********************************************************
**********************************************************
      real function f(x,nfunc)
      real x
      integer nfunc
c     f(x)=max(x,1-x)
      if (nfunc .eq. 1) then
	if (x .le. 0.5) then
	  f=x
	else
	  f=1.0-x
	endif
c     f(x)=1
      else if (nfunc .eq. 2) then
	f=1.0
      end if
      end
